// Program to read Aportis format files on the Agenda VR3
// by $Author: agenda $. Current version $Revision: 1.40 $, $Date: 2001/11/07 21:33:58 $.
// This is a $State: Exp $ version.
// Baseline is $Name: V1_2 $
//
// $Log: reader.cxx,v $
// Revision 1.40  2001/11/07 21:33:58  agenda
// Final tweaks for v1.2
//
// Revision 1.39  2001/10/21 14:02:52  agenda
// Added deactivation of menu items etc.
//
// Revision 1.38  2001/10/16 22:12:45  agenda
// Added "in-memory" autogenned bookmarks.
//
// Revision 1.37  2001/10/16 21:06:21  agenda
// Changes title to match currently loaded file.
//
// Revision 1.36  2001/10/10 19:55:20  agenda
// First release candidate with background autogenning.
//
// Revision 1.35  2001/10/10 19:37:40  agenda
// Background autogenning seems to be working - now to tidy it up!
//
// Revision 1.34  2001/10/04 21:29:53  agenda
// First round of "slow and small" bug fixes
//
// Revision 1.33  2001/10/03 19:19:17  agenda
// Slow but small background auto-genning
//
// Revision 1.32  2001/10/02 21:41:11  agenda
// Background autogen
//
// Revision 1.31  2001/09/08 23:08:45  agenda
// Added tight spacing option
//
// Revision 1.30  2001/09/07 14:56:26  agenda
// Improvements to window placement etc.
//
// Revision 1.29  2001/09/06 00:03:41  agenda
// Removed FP version of fl_width, put CBars info info dialogue and
// made the buffdoc and myoutput objects a little more self-contained.
//
// Revision 1.28  2001/09/05 02:22:26  agenda
// Converted myoutput to a widget, added progress bars and some minor bug fixes.
//
// Revision 1.27  2001/09/01 14:30:13  agenda
// Check-in for actual beta-release 1.0a
//
// Revision 1.26  2001/08/30 00:07:06  agenda
// Added filesize info and prepared for next release
//
// Revision 1.25  2001/08/28 10:31:13  agenda
// Version 1.0 beta A
//
// Revision 1.24  2001/08/16 20:23:45  agenda
// First working version with bookmarks
//
// Revision 1.23  2001/08/05 19:14:06  agenda
// Started adding in bookmarks proper
//
// Revision 1.21  2001/07/04 21:16:29  agenda
// Ready for initial release
//
// Revision 1.20  2001/06/19 21:24:57  agenda
// First release candidate
//
// Revision 1.19  2001/06/13 21:06:43  agenda
// Now works reasonably well and quite quickly too!
//
// Revision 1.18  2001/06/10 21:29:53  agenda
// Replaced multiline output with direct drawing of text
//
// Revision 1.17  2001/06/07 23:34:00  agenda
// First working(?) version with Flek and flpda.
//
// Revision 1.16  2001/06/07 21:54:51  agenda
// Changed over to use text rather than Sixpack compression as the third data type.
//
// Revision 1.15  2001/06/07 21:41:47  agenda
// Made a bit more robust!
//
// Revision 1.14  2001/06/07 21:30:00  agenda
// First working version
//
// Revision 1.16  2001/06/01 12:09:55  agenda
// Version which works with ppm
//
// Revision 1.15  2001/05/28 05:09:10  agenda
// Done some restructuring so that plugging in other decompression schemes should be easier.
// Changed the page-up code to be more efficient and more accurate - though it may get messy if there is no formatting
// in the input file as it looks for paragraphs to regain synchronisation.
//
// Still to do:
//   User defined bookmarks
//   Allow some overlap between pages
//   Find
//   Tidy up the page-up code to use a circular array of pointers and a single
//   char* rather than a circular array of char*'s
//   Tidy up the page-up code in general
//
// Revision 1.14  2001/05/26 10:44:43  agenda
// Check-in just prior to some restructuring
//
// Revision 1.13  2001/05/11 05:07:59  agenda
// First attempt to do bookmarks. Works on workstation but fails on Agenda.
//
// Revision 1.12  2001/05/11 03:25:17  agenda
// Corrected a bug when a match overlaps the current position. Added two optional command line arguments to open a particular file at a particuler location at startup.
//
// Revision 1.11  2001/04/29 22:21:36  agenda
// Added a little bit of book-marking but I couldn't get the dbm stuff
// on the Agenda to work.
// Added magnify and reduce and embolden.
//
// Revision 1.10  2001/04/22 21:38:58  agenda
// Removed obsolete tBuf class, made buffers a more realistic size
// and a bit of tidying up.
//
// Revision 1.9  2001/04/22 14:42:47  agenda
// Page up and Page down now working, though page up is approximate and not
// particularly efficient. Most of the stuff is there to allow bookmarks
// but the actual GUI and storage for bookmarks is still TBD.
//
// Currently can use Alt-f to query/set the current position so bookmarks
// can be managed manually, for now.
//
// Revision 1.8  2001/04/21 20:38:00  agenda
// Added some stuff to go to a particular point in the file.
// To get it working properly requires some major changes to
// the decompression stuff so this check-in is just so I
// can revert if I mess up the next set of changes.
//
// Revision 1.7  2001/04/21 08:44:28  agenda
// Tidied up page down a bit.
// Other stuff to do which I left out of my last check-in is to get find working
// (or maybe just remove it)
//
// Revision 1.6  2001/04/21 08:07:00  agenda
// First useable version with buffering.
// Still to do:
//   Enable page-up
//   Bookmarks
//   Allow some overlap between pages
// Currently each page has no text in common with the previous page.
// Normally you would expect the first line of a new page to be the same
// as the last line of the next page. I happen to quite like how it
// (accidentally) works now so I may not change it.
//
// Revision 1.4  2001/04/14 23:05:03  agenda
// Tidied up header and clarified where other headers are quoted.
//
// Revision 1.3  2001/04/14 22:53:16  agenda
// Begun adding in version control info, etc.
//
// Revision 1.2  2001/04/14 22:40:44  agenda
// Improved handling of page-up/down events
//
// Revision 1.1  2001/04/14 21:44:23  agenda
// Initial revision
// ============================================================================
//
static char rcsid[] = "$Id: reader.cxx,v 1.40 2001/11/07 21:33:58 agenda Exp $";
// Decompression based (very heavily) on:
//
// ------------- Begin quoted header -------------
// MakeDoc
// version 2
//
// Compresses text files into a format that is ready to export to a Pilot
// and work with Rick Bram's PilotDOC reader.
//
// Copyright (C) Pat Beirne, 2000.
// Distributable under the GNU General Public License Version 2 or later.
//
// ver 0.6 enforce 31 char limit on database names
// ver 0.7 change header and record0 to structs
// ver 2.0 added category control on the command line
//		changed extensions from .prc to .pdb
// -------------- End quoted header --------------
//
// Remainder based (more loosely) on
//
// ------------- Begin quoted header -------------
// A simple text editor program for the Fast Light Tool Kit (FLTK).
//
// This program is described in Chapter 4 of the FLTK Programmer's Guide.
//
// Copyright 1998-2000 by Bill Spitzak and others.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Library General Public License for more details.
//
// You should have received a copy of the GNU Library General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA.
//
// Please report all bugs and problems to "fltk-bugs@fltk.org".
// -------------- End quoted header --------------
//
// Note that it is unlikely that any bug found in this program
// will be relevant to the fltk organisation as it will almost
// certainly be one of mine.

#include <stdio.h>			// Standard library files
#ifdef _SLOW
#include <unistd.h>
#endif
#include <stdlib.h>
#include <string.h>
#include <errno.h>

//#include <string>

#ifdef AGENDA
#include <flpda/Widget_Factory.h>
#include <Flek/Fl_App_Window.H>
#endif

#include <FL/Fl.H>			// Main FLTK header file
#include <FL/Fl_Group.H>		// Fl_Group header file
#include <FL/Fl_Window.H>		// Fl_Window header file
#include <FL/fl_ask.H>			// FLTK convenience functions
#if defined(RAW) || !defined(AGENDA)
#include <FL/fl_file_chooser.H>
#else
//#include "fl_ag_file_chooser.H"		// FLTK file chooser
#include "my_fl_file_chooser.H"		// FLTK file chooser
#endif
#include <FL/Fl_Menu_Bar.H>		// Fl_Menu_Bar header file
#include <FL/Fl_Input.H>		// Fl_Input header file
#ifdef AGENDA
#include <flpda/Fl_Paged_Select_Browser.H>	// Fl_Multiline_Input header file
#else
#include <FL/Fl_Select_Browser.H>	// Fl_Multiline_Input header file
#endif
#include <FL/Fl_Multiline_Output.H>	// Fl_Multiline_Input header file
#include <FL/Fl_Button.H>		// Fl_Button header file
#include <FL/Fl_Return_Button.H>	// Fl_Return_Button header file
#include <FL/Fl_Int_Input.H>
#include <FL/fl_draw.H>
#include <FL/filename.H>
#include <sys/stat.h>
#include <fcntl.h>
#ifdef USEBKMKS
#include "Queue.h"
#ifdef WS
#include <db1/db.h>
#else
#include <db.h>
#endif
#endif
#include "CFilter.h"
#include "CBuffer.h"

#include "CBar.h"

//#include "unistd.h"

#ifdef USEBKMKS
#define LASTFILE "/////lastfileopened/////"
#define PREFRNCE "/////preferences/////"
#endif

#ifdef USEBKMKS
void setbkmkkey(const char *file, const char *mark, DBT* key);
DB* bkmks;
#endif

typedef unsigned char byte;
typedef unsigned long DWORD;
typedef unsigned short WORD;

char               filename[1024];
char               search[256];

#include "Aportis.h"
#ifndef SMALL
#include "ppm_expander.h"
#endif
//#include "Sixpack.h"
#include "Text.h"

#ifdef AGENDA
void centre_window(Fl_Window* p)
{
  p->position((Fl::w()-p->w()) >> 1, (Fl::h()-p->h()) >> 1);
}
#endif

void Cancel_cb(Fl_Widget *widget, void *data)
{
  widget->window()->hide();
}

class BuffDoc {
  CBuffer lastword;
  bool lastispara;
  CExpander* exp;
  CFilterChain* filt;
public:
  ~BuffDoc()
  {
    delete filt;
    delete exp;
  }
  BuffDoc()
  {
    exp = NULL;
    filt = NULL;
  }
  void setfilter(CFilterChain* _f)
  {
    if (filt != NULL) delete filt;
    filt = _f;
    filt->setsource(exp);
  }
  CList<Bkmk>* getbkmklist() { return exp->getbkmklist(); }
  bool hasrandomaccess() { return (exp == NULL) ? false : exp->hasrandomaccess(); }
  bool iseol() { return (lastword[0] == '\0'); }
  int openfile(char *src)
  {
    //    printf("Trying aportis %x\n",exp);
    if (exp != NULL) delete exp;
    exp = new Aportis;
    lastword[0] = '\0';
    lastispara = false;
    //    printf("Calling openfile\n");
    int ret = exp->openfile(src);
    //    printf("Called openfile\n");
    if (ret == -1)
      {
	// fl_alert("problem opening source file:%s",src);
	//	printf("problem opening source file\n");
	delete exp;
	exp = NULL;
	return ret;
      }
    if (ret == -2)
      {
	//	printf("file contains %x, %x",head.dwCreator, head.dwType);
	delete exp;
#ifndef SMALL
	//	printf("Trying ppms\n");
	exp = new ppm_expander;
	ret = exp->openfile(src);
	if (ret != 0) {
	  delete exp;
	  exp = new Text;
	  //	  printf("Trying text\n");
	  ret = exp->openfile(src);
	}
#else
	exp = new Text;
	ret = exp->openfile(src);
#endif
      }
    if (ret != 0)
      {
	fl_alert("WARNING: unknown file compression type:%d",ret);
      }
    lastword[0] = '\0';
    lastispara = false;
    exp->locate(0);
    filt->setsource(exp);
    return 0;
  }
  int getch() { return (exp == NULL) ? EOF : filt->getch(); }
  unsigned int locate() { return (exp == NULL) ? 0 : exp->locate()-strlen(lastword.data()); }
  void locate(unsigned int n);
  bool getline(CBuffer* buff, int w);
  void sizes(unsigned long& fs, unsigned long& ts) { exp->sizes(fs,ts); }
  int getpara(CBuffer& buff)
  {
    int ch, i = 0;
    while ((ch = getch()) != 10 && ch != EOF) buff[i++] = ch;
    buff[i] = '\0';
    if (i == 0 && ch == EOF) return -1;
    return i;
  }
};

void BuffDoc::locate(unsigned int n)
{
    
  lastword[0] = '\0';
  lastispara = false;
  //    char linebuf[1024];
  if (exp != NULL) exp->locate(n);
  //    while (exp.locate() < n && getline(linebuf,w));
    
}

bool BuffDoc::getline(CBuffer* buff, int w)
{
  int len = 0, ch, lastcheck = 0;
  *buff = lastword.data();
  lastcheck = len = buff->length();
  unsigned int slen = fl_widthi(buff->data(),len);
  if (slen > w)
    {
      for ( ; len > 0; len--)
	{
	  if (fl_widthi(buff->data(),len) < w) break;
	}
      lastword = buff->data() + len - 1;
      (*buff)[len-1] = '-';
      (*buff)[len] = '\0';
      return true;
    }
  if (lastispara)
    {
      lastispara = false;
      lastword[0] = '\0';
      len = buff->length();
      while (fl_widthi(buff->data(),len) > w) len--;
      (*buff)[len] = '\0';
      return true;
    }
  lastispara = false;
  while (slen < w)
    {
      lastcheck = len;
      while ((ch = getch()) != ' ' && ch != '\012' && ch != EOF)
	{
	  (*buff)[len++] = ch;
	}
      slen = fl_widthi(buff->data(),len);
      (*buff)[len++] = ' ';
      if (slen < w && ch != ' ') {
	lastcheck = len;
	break;
      }
      lastispara = (ch == '\012');
    }
  (*buff)[len] = '\0';
  lastword = buff->data()+lastcheck;
  (*buff)[(lastcheck > 0) ? lastcheck-1:lastcheck] = '\0';
  return (ch != EOF);
}

class myoutput : public Fl_Widget {
  Fl_Font m_textfont;
  int m_textsize;
  CBuffer** textarray;
  unsigned int numlines;
public:
  int getch() { return buffdoc.getch(); }
  bool tight;
  myoutput(int x, int y, int w, int h, const char *label=0) : Fl_Widget(x,y,w,h,label), textarray(NULL), numlines(0), tight(true) {}
  virtual void draw();
  virtual int handle(int event);
  bool load_file(char *newfile, unsigned int lcn=0);
  BuffDoc buffdoc;
  CList<Bkmk>* getbkmklist() { return buffdoc.getbkmklist(); }
  bool locate(unsigned int n);
  void jumpto(unsigned int n) { buffdoc.locate(n); }
  unsigned int locate() { return buffdoc.locate(); }
  unsigned int pagelocate() { return pagepos; }
  unsigned int pagepos, mylastpos;
  void setfilter(CFilterChain *f) { buffdoc.setfilter(f); locate(pagepos); }
  void restore() { jumpto(mylastpos); }
  void dopageup();
  void dopagedn()
  {
    if (locate() != mylastpos) jumpto(mylastpos);
    if (fillbuffer())
      {
	redraw();
	Fl::flush();
      }
  }
  virtual void resize(int X, int Y, int W, int H)
  {
    buffdoc.locate(pagepos);
    fillbuffer();
    Fl_Widget::resize(X,Y,W,H);
  }
  //  bool bold;
  int textsize() { return m_textsize; }
  void textsize(int ts) { m_textsize = ts; }
  Fl_Font textfont() { return m_textfont; }
  void textfont(Fl_Font tf) { m_textfont = tf; }
  bool fillbuffer();
  int screenlines();
  void sizes(unsigned long& fs, unsigned long& ts) { buffdoc.sizes(fs,ts); }
};

bool myoutput::locate(unsigned int n) {
  //printf("Locate\n");
  buffdoc.locate(n);
  fillbuffer();
  redraw();
}

int myoutput::screenlines()
{
  fl_font(textfont(),textsize());
  int linespacing = (tight) ? fl_height() : fl_height()+fl_descent();
  //  int delta = (h()-fl_height())/(fl_height());
  //int delta = (parent()->h()-tb->h())/(fl_height());
#ifdef AGENDA
  //  return (parent()->h()-fl_descent())/(fl_height());
  return (h()-fl_descent())/(linespacing);
#else
  return (h()-fl_descent())/(linespacing);
#endif
};

bool myoutput::fillbuffer() {
  //printf("Fillbuffer\n");
  int ch;
  bool ret = false;
  int delta = screenlines();
  if (delta != numlines)
    {
      if (textarray != NULL)
	{
	  for (int i = 0; i < numlines; i++) delete textarray[i];
	  delete [] textarray;
	}
      numlines = delta;
      textarray = new CBuffer*[numlines];
      for (int i = 0; i < numlines; i++) textarray[i] = new CBuffer;
    }
  unsigned int oldpagepos = pagepos;
  pagepos = locate();
  for (int i = 0; i < delta; i++)
    {
      ch = buffdoc.getline(textarray[i],w());
      //      if (ch == EOF) {
      if (!ch)
	{
	  if (i == 0)
	    {
	      pagepos = oldpagepos;
	      return false;
	    }
	  else
	    {
	      ret = true;
	      for (int j = i; j < delta; j++) (*(textarray[j]))[0] = '\0';
	      break;
	    }
	}
      if (ch == '\012') ret = true;
    }
  mylastpos = locate();
  return true;
}

void myoutput::dopageup()
{
  fl_font(textfont(),textsize());
  CBuffer** buff = textarray;
  unsigned int *loc = new unsigned int[numlines];
  int cbptr = 0;
  if (locate() != mylastpos) jumpto(mylastpos);
  unsigned int target = pagelocate();
  if (buffdoc.hasrandomaccess())
    {
      unsigned int delta = locate()-pagelocate();
      if (delta < 64) delta = 64;
      do
	{
	  delta <<= 1;
	  if (delta >= target)
	    {
	      delta = target;
	      jumpto(0);
	      for (int i = 0; i < numlines; i++)
		{
		  loc[i] = locate();
		  buffdoc.getline(buff[i],w());
		}
	      break;
	    }
	  jumpto(target-delta);
	  do
	    {
	      buffdoc.getline(buff[0],w());
#ifdef WS
	      //printf("Trying:%s\n",buff[0]);
#endif
	      if (locate() > target) continue;
	    }
	  while (!buffdoc.iseol());
	  for (int i = 0; i < numlines; i++)
	    {
	      loc[i] = locate();
	      buffdoc.getline(buff[i],w());
#ifdef WS
	      //printf("Filling:%s\n",buff[i]);
#endif
	    }
	}
      while (locate() >= target && delta < 4096);
#ifdef WS
      //printf("Delta:%u\n",delta);
#endif
    }
  else
    {
      jumpto(0);
      for (int i = 0; i < numlines; i++)
	{
	  loc[i] = locate();
	  buffdoc.getline(buff[i],w());
	}
    }
  cbptr = 0;
  while (locate() < target)
    {
      loc[cbptr] = locate();
      buffdoc.getline(buff[cbptr], w());
#ifdef WS
      //printf("Adding:%s\n",buff[cbptr]->data());
#endif
      cbptr = (cbptr+1) % numlines;
    }
  pagepos = loc[cbptr];
  textarray = new CBuffer*[numlines];
  for (int i = 0; i < numlines; i++)
    {
      textarray[i] = buff[(cbptr+i)%numlines];
    }
  delete [] buff;
  delete [] loc;
  mylastpos = locate();
  redraw();
  Fl::flush();
}

int myoutput::handle(int event)
{
  //  printf("Event:%d\n",event);
  if (event == FL_FOCUS || event == FL_UNFOCUS)
    {
      return 1;
    }
  else if (event == FL_KEYBOARD)
    {
      switch (Fl::event_key())
        {
        case FL_Page_Up:
	  {
	    dopageup();
	  }
	  return 1;
        case FL_Page_Down:
	  {
	    dopagedn();
	  }
	  return 1;
	}
    }
  else if (event == FL_PUSH)
    {
      if (Fl::event_y() < h()/2)
	{
	  dopageup();
	}
      else
	{
	  dopagedn();
	}
      return 1;
    }
  return Fl_Widget::handle(event);
}

bool myoutput::load_file(char *newfile, unsigned int lcn)
{
  bool bRC = false;
  if (buffdoc.openfile(newfile) == 0)
    {
      bRC = true;
      locate(lcn);
    }
  redraw();
  return bRC;
}

#ifdef USEBKMKS
void setbkmkkey(const char *file, const char *mark, DBT* key)
{
  static char buffer[2048];
  int flen = strlen(file);
  int clen = strlen(mark);
  //  buffer = new char[flen+1+clen];
  strcpy(buffer,file);
  memcpy(buffer+flen+1,mark,clen);
  key->data = buffer;
  key->size = flen+1+clen;
}

void getdefaultbkmk(char *file, DBT* key)
{
  setbkmkkey(file,"CURRENT",key);
}
#endif

void myoutput::draw() {
  fl_color(7);
  fl_rectf(x(),y(),x()+w(),y()+h());
  fl_color(0);
  int sl = screenlines();
  if (sl < numlines)
    {
      size_t offset = 0;
      CBuffer** nta = new CBuffer*[sl];
      for (int i = 0; i < sl; i++)
	{
	  nta[i] = textarray[i];
	  offset += nta[i]->length()+1;
	}
      for (int i = sl; i < numlines; i++) delete textarray[i];
      delete [] textarray;
      textarray = nta;
      numlines = sl;
      jumpto(mylastpos = pagepos+offset);
    }
  if (sl > numlines)
    {
      size_t offset = 0;
      CBuffer** nta = new CBuffer*[sl];
      for (int i = 0; i < numlines; i++)
	{
	  nta[i] = textarray[i];
	  offset += nta[i]->length();
	}
      if (locate() != mylastpos) jumpto(mylastpos);
      for (int i = numlines; i < sl; i++)
	{
	  nta[i] = new CBuffer;
	  buffdoc.getline(nta[i],w());
	}
      mylastpos = locate();
      delete [] textarray;
      textarray = nta;
      numlines = sl;
    }
  int ypos = (tight) ? y() : y()-fl_descent();
  int linespacing = (tight) ? fl_height() : fl_height()+fl_descent();
  for (int i = 0; i < numlines; i++)
    {
      fl_draw(textarray[i]->data(),x(),ypos += linespacing);
    }
}

struct prefs_t
{
  int textsize;
  Fl_Font textfont;
  bool bold;
  bool bstripcr, bunindent, brepara, bdblspce, btight;
  unsigned char bindenter;
};

class reader_ui
{
  Fl_Menu_Item *mi_zoomin, *mi_zoomout, *mi_savebkmk, *mi_autogen;
  CQueue<Bkmk>* pBkmklist;
  bool bold;
  char g_filename[1024];
  myoutput *input;
#ifdef AGENDA
  Fl_App_Window          *window;
#else
  Fl_Window *window;
#endif
#ifdef USEBKMKS
  unsigned int initlocation(char *file);
  unsigned int mark;
#endif
  Fl_Int_Input *locn;
#ifdef USEBKMKS
  Fl_Window* bkmkwindow;
  bool bkmkfind;
#endif
  bool bstripcr, bunindent, brepara, bdblspce, btight;
  unsigned char bindenter;
#ifdef USEBKMKS
#ifdef AGENDA
  Fl_Paged_Select_Browser* bkmkbrowser;
#else
  Fl_Select_Browser* bkmkbrowser;
#endif
#endif
  void setfilter()
  {
    input->setfilter(GetFilter());
  }
  CFilterChain* GetFilter()
  {
    CFilterChain * filt = new CFilterChain;
    if (bstripcr) filt->addfilter(new stripcr);
    if (bunindent) filt->addfilter(new unindent);
    if (brepara) filt->addfilter(new repara);
    if (bindenter) filt->addfilter(new indenter(bindenter));
    if (bdblspce) filt->addfilter(new dblspce);
    return filt;
  }
  //Fl_Menu_Bar        *menubar;
#ifdef USEBKMKS
  static void save_cb(Fl_Widget *widget, void *_data) {
    myoutput* input = (((reader_ui*)_data)->input);
    if (((reader_ui*)_data)->g_filename[0] != '\0')
      {
	const char *bkmark = fl_input("Bookmark name", "CURRENT");
	if (bkmark != NULL)
	  {
	    DBT key, data;
	    memset(&key,0,sizeof(key));
	    memset(&data,0,sizeof(data));
	    setbkmkkey(((reader_ui*)_data)->g_filename,bkmark,&key);
	    unsigned int lcn = input->pagelocate();
	    data.data = (char*)&lcn;
	    data.size = sizeof(lcn);
	    bkmks->put(bkmks, &key, &data, 0);
	    //printf("sync:%x\n",bkmks->sync);
	    bkmks->sync(bkmks, 0);
	  }
      }
  }
#endif

#ifdef USEBKMKS
  static void mark_cb(Fl_Widget *widget, void *data)
  {
    ((reader_ui*)data)->mark = ((reader_ui*)data)->input->pagelocate();
  }
  static void copy_cb(Fl_Widget *widget, void *data)
  {
    ((reader_ui*)data)->copy(((reader_ui*)data)->mark, ((reader_ui*)data)->input->locate());
  }
  void copy(unsigned int start, unsigned int end)
  {
    unsigned int current = input->locate();
    if (start > end)
      {
	unsigned int temp = start;
	start = end;
	end = temp;
      }
    unsigned int len = end-start;
    input->jumpto(start);
    CBuffer buff;
    for (int i = 0; i < len; i++)
	buff[i] = input->getch();
    Fl::selection(*input, buff.data(), len);
    input->jumpto(current);
  }

  static void autogenbkmk_cb(Fl_Widget *widget, void *data)
  {
    ((reader_ui*)data)->mi_autogen->deactivate();
    static Fl_Window* win = NULL;
    static CBar* bar = NULL;
    if (win == NULL)
      {
	win = new Fl_Window(133,228,27,12);
	//	win->resizable(win);
	bar = new CBar(1,1,25,10);
	win->border(0);
	win->box(FL_UP_BOX);
	win->end();
      }
    bar->value(0);
    win->show();
    ((reader_ui*)data)->do_autogen(bar);
    ((reader_ui*)data)->mi_autogen->activate();
    win->hide();
  }
  void do_autogen(CBar* pbar)
  {
    CBuffer buff;
    if (pBkmklist != NULL) delete pBkmklist;
    pBkmklist = new CQueue<Bkmk>;
    unsigned long fs,ts;
    input->sizes(fs,ts);
    int ch = 0;
    if (pbar->value(0))
      {
#ifdef _SLOW
	sleep(1);
#endif
	Fl::check();
      }
    input->jumpto(0);
    int i = 0;
    while (i >= 0)
      {
	unsigned int lcn = input->locate();
	if (pbar->value(((100*lcn)/ts)))
	  {
#ifdef _SLOW
	    sleep(1);
#endif
	    unsigned long pos = input->locate();
	    Fl::check();
	    if (input->locate() != pos) input->jumpto(pos);
	  }
	i = input->buffdoc.getpara(buff);
	if (i > 0)
	  {
	    i--;
	    while (i > 0 && (ch = buff[i]) == ' ') i--;
	    if ((('a' <= ch && ch <= 'z') || ('A' <= ch && ch <= 'Z') || ch == ' ') && i < 128)
	      {
		/*
		DBT key, data;
		memset(&key,0,sizeof(key));
		memset(&data,0,sizeof(data));
		setbkmkkey(g_filename,buff.data(),&key);
		data.data = (char*)&lcn;
		data.size = sizeof(lcn);
		bkmks->put(bkmks, &key, &data, 0);
		*/
		pBkmklist->push_back(Bkmk(buff.data(),lcn));
	      }
	    //	    else printf("%x:%c\n",ch,ch);
	  }
      }
    bkmks->sync(bkmks, 0);
    mi_savebkmk->activate();
    input->restore();
  }

  static void clrbkmk_cb(Fl_Widget *widget, void *_data)
  {
    CQueue<Bkmk>* t = ((reader_ui*)_data)->pBkmklist;
    if (t != NULL)
      {
	delete t;
	((reader_ui*)_data)->pBkmklist = NULL;
      }
    else
      {
	DBT key, data;
	memset(&key,0,sizeof(key));
	memset(&data,0,sizeof(data));
	key.data = ((reader_ui *)_data)->g_filename;
	key.size = strlen(((reader_ui *)_data)->g_filename);
	bkmks->seq(bkmks,&key,&data,R_CURSOR);
	while (strcmp(((reader_ui *)_data)->g_filename,(char*)key.data)==0)
	  {
	    bkmks->del(bkmks,&key,R_CURSOR);
	    if (bkmks->seq(bkmks,&key,&data,R_NEXT) != 0) break;
	  }
	bkmks->sync(bkmks,0);
      }
  }

  static void savebkmk_cb(Fl_Widget *widget, void *_data)
  {
    ((reader_ui*)_data)->mi_savebkmk->deactivate();
    CQueue<Bkmk>* t = ((reader_ui*)_data)->pBkmklist;
    if (t != NULL)
      {
	/*
	  for (CList<Bkmk>::iterator i = t->begin(); i != t->end(); i++)
	  {
	  DBT key, data;
	  memset(&key,0,sizeof(key));
	  memset(&data,0,sizeof(data));
	  setbkmkkey(((reader_ui*)_data)->g_filename,i->name(),&key);
	  unsigned int lcn = i->value();
	  data.data = (char*)&lcn;
	  data.size = sizeof(lcn);
	  bkmks->put(bkmks, &key, &data, 0);
	  }
	*/
	while (!t->empty())
	  {
	    Bkmk b = t->pop();
	    DBT key, data;
	    memset(&key,0,sizeof(key));
	    memset(&data,0,sizeof(data));
	    setbkmkkey(((reader_ui*)_data)->g_filename,b.name(),&key);
	    unsigned int lcn = b.value();
	    data.data = (char*)&lcn;
	    data.size = sizeof(lcn);
	    bkmks->put(bkmks, &key, &data, 0);
	  }

	bkmks->sync(bkmks,0);
	delete t;
	((reader_ui*)_data)->pBkmklist = NULL;
      }
  }

  static void tidybkmk_cb(Fl_Widget *widget, void *_data)
  {
    DBT key, data;
    memset(&key,0,sizeof(key));
    memset(&data,0,sizeof(data));
    bkmks->seq(bkmks,&key,&data,R_FIRST);
    do
      {
	struct stat buf;
	char buffer[1024];
	memcpy(buffer, key.data, key.size);
	buffer[key.size] = '\0';
	if (strcmp(buffer,LASTFILE) == 0) continue;
	if (strcmp(buffer,PREFRNCE) == 0) continue;
	if (stat(buffer,&buf) == -1 && errno == ENOENT)
	  {
	    bkmks->del(bkmks,&key,R_CURSOR);
	  }
      }
    while (bkmks->seq(bkmks,&key,&data,R_NEXT) == 0);
    bkmks->sync(bkmks,0);
  }
#endif

#ifdef USEBKMKS
  static void save_current(Fl_Widget *widget, void *_data) {
    myoutput* input = (((reader_ui*)_data)->input);
    if (((reader_ui *)_data)->g_filename[0] != '\0')
      {
	DBT key, data;
	memset(&key,0,sizeof(key));
	memset(&data,0,sizeof(data));
	getdefaultbkmk(((reader_ui *)_data)->g_filename, &key);
	unsigned int lcn = input->pagelocate();
	data.data = (char*)&lcn;
	data.size = sizeof(lcn);
	bkmks->put(bkmks, &key, &data, 0);
	bkmks->sync(bkmks, 0);
      }
  }
#endif

  static Fl_Font rotate_font(Fl_Font font)
  {
    switch (font)
      {
      case FL_HELVETICA: return FL_TIMES;
      case FL_TIMES: return FL_COURIER;
      case FL_COURIER: return FL_HELVETICA;
      case FL_HELVETICA_BOLD: return FL_TIMES_BOLD;
      case FL_TIMES_BOLD: return FL_COURIER_BOLD;
      case FL_COURIER_BOLD: return FL_HELVETICA_BOLD;
      }
  }

  static Fl_Font embolden(Fl_Font font)
  {
    switch (font)
      {
      case FL_HELVETICA: return FL_HELVETICA_BOLD;
      case FL_TIMES: return FL_TIMES_BOLD;
      case FL_COURIER: return FL_COURIER_BOLD;
      }
  }

  static Fl_Font enlighten(Fl_Font font)
  {
    switch (font)
      {
      case FL_HELVETICA_BOLD: return FL_HELVETICA;
      case FL_TIMES_BOLD: return FL_TIMES;
      case FL_COURIER_BOLD: return FL_COURIER;
      }
  }

  static void bld_cb(Fl_Widget *widget, void *data) {
    myoutput* input = (((reader_ui*)data)->input);
    ((reader_ui*)data)->bold = !((reader_ui*)data)->bold;
    input->textfont(((reader_ui*)data)->bold ? embolden(input->textfont()) : enlighten(input->textfont()));
    input->locate(input->pagelocate());
    input->redraw();
  }

  static void tight_cb(Fl_Widget *widget, void *data) {
    myoutput* input = (((reader_ui*)data)->input);
    ((reader_ui*)data)->btight = !((reader_ui*)data)->btight;
    input->tight = ((reader_ui*)data)->btight;
    input->redraw();
  }

  static void fnt_cb(Fl_Widget *widget, void *data) {
    myoutput* input = (((reader_ui*)data)->input);
    input->textfont(rotate_font(input->textfont()));
    input->locate(input->pagelocate());
    input->redraw();
  }

  static void mag_cb(Fl_Widget *widget, void *data) {
    myoutput* input = (((reader_ui*)data)->input);
    int * sizep;
    int n = Fl::get_font_sizes(input->textfont(),sizep);
    if (n == 0) return;
    ((reader_ui*)data)->mi_zoomout->activate();
    if (sizep[0] == 0 && n <= 1)
      {
	input->textsize(input->textsize()+1);
      }
    else
      {
	int i;
	for (i = 0; i < n-1; i++)
	  {
	    if (sizep[i] > input->textsize()) break;
	  }
	input->textsize(sizep[i]);
	if (i == n-1)
	  {
	    ((reader_ui*)data)->mi_zoomin->deactivate();
	  }
      }
    input->locate(input->pagelocate());
    input->redraw();
  }

  static void red_cb(Fl_Widget *widget, void *data) {
    myoutput* input = (((reader_ui*)data)->input);
    int * sizep;
    int n = Fl::get_font_sizes(input->textfont(),sizep);
    if (n == 0) return;
    ((reader_ui*)data)->mi_zoomin->activate();
    if (sizep[0] == 0 && n <= 1)
      {
	input->textsize(input->textsize()-1);
      }
    else
      {
	int i;
	for (i = n-1; i >= 0; i--)
	  {
	    if (sizep[i] < input->textsize()) break;
	  }
	if (i == 0 || (i == 1 && sizep[0] == 0))
	  {
	    ((reader_ui*)data)->mi_zoomout->deactivate();
	  }
	if (sizep[i] == 0) i++;
	input->textsize(sizep[i]);
      }

    input->locate(input->pagelocate());
    input->redraw();
  }

  static void jump_cb(Fl_Widget *widget, void *_data) {
    char vlu[40];
    static Fl_Window *jumptowin = NULL;

    if (jumptowin == NULL)
      {
	jumptowin = new Fl_Window(100,45);
#ifdef AGENDA
	centre_window(jumptowin);
#endif
	jumptowin->border(0);
	jumptowin->box(FL_UP_BOX);
	((reader_ui *)_data)->locn = new Fl_Int_Input(35,5,50,14,"Place");
	((reader_ui *)_data)->locn->textsize(10);
	((reader_ui *)_data)->locn->labelsize(10);
	Fl_Button *jump_accept = new Fl_Button(10,25,35,15,"Okay");
	jump_accept->labelsize(10);
	jump_accept->callback(jumpAccept_cb,_data);
	Fl_Button *jump_cancel = new Fl_Button(55,25,35,15,"Cancel");
	jump_cancel->labelsize(10);
	jump_cancel->callback(Cancel_cb,_data);
	jumptowin->end();
	jumptowin->set_modal();
      }
    sprintf(vlu, "%d", ((reader_ui *)_data)->input->pagelocate());
    ((reader_ui *)_data)->locn->value(vlu);
    jumptowin->show();
  }

  static void jumpAccept_cb(Fl_Widget *widget, void *data)
  {
    widget->window()->hide();
    ((reader_ui *)data)->input->locate(atoi(((reader_ui *)data)->locn->value()));
  }

  static void find_cb(Fl_Widget *widget, void *data) {
    const char *val;
    myoutput* input = (((reader_ui*)data)->input);

    //sprintf(search, "%d", input->pagelocate());
    val = fl_input("Search String:", search);

    if (val != NULL) {
      // User entered a string - go find it!
      strcpy(search, val);
      find2_cb(widget, data);
    }
  }

  static void find2_cb(Fl_Widget *widget, void *data) {
    const char *val, *found;
    int pos;
    myoutput* input = (((reader_ui*)data)->input);

    if (search[0] == '\0') {
      // Search string is blank; get a new one...
      find_cb(widget,data);
      return;
    }

    if (found != NULL) {
      CBuffer test;
      //      input->jumpto(input->locate());
      size_t pos;
      size_t start = input->pagelocate();
      input->jumpto(start);
      input->buffdoc.getline(&test,input->w());
      do
	{
	  pos = input->locate();
	  if (!input->buffdoc.getline(&test,input->w()))
	    {
	      fl_alert("No occurrences of \'%s\' found!", search);
	      pos = start;
	      break;
	    }
	  //	  printf("%d:%s\n",pos,test.data());
	}
      while (strstr(test.data(),search) == NULL);
      input->locate(pos);
    }
  }

  static void open_cb(Fl_Widget *widget, void *data) {
    char *newfile;
    reader_ui* reader = (reader_ui*)data;
    myoutput* input = reader->input;
#if defined(RAW) || !defined(AGENDA)
    newfile = fl_file_chooser("Open File?", "*", filename);
#else
    //newfile = fl_ag_file_chooser("Open File?", "*", filename);
    newfile = my_fl_file_chooser("Open File?", "*", filename);
#endif
    if (newfile != NULL) reader->load_file(newfile);
  }

  static void quit_cb(Fl_Widget *widget, void *_data) {
    reader_ui* reader = (reader_ui*)_data;
#ifdef USEBKMKS
    save_current(widget, _data);
#endif
    reader->setprefs();
#ifdef USEBKMKS
    DBT key, data;
    memset(&key,0,sizeof(key));
    memset(&data,0,sizeof(data));
    char buffer[1024];
    strcpy(buffer,LASTFILE);
    key.data = buffer;
    key.size = strlen(buffer);
    data.data = ((reader_ui *)_data)->g_filename;
    data.size = strlen(((reader_ui *)_data)->g_filename);
    bkmks->put(bkmks,&key,&data,0);
    bkmks->close(bkmks);
#endif
    reader->window->hide();
  }
  reader_ui(const reader_ui&);
public:

#ifdef USEBKMKS
  void setupbkmks()
  {
#ifdef AGENDA
    bkmkwindow = new Fl_Window(160,240,"Choose Bookmark");
    //    bkmkwindow->box(FL_NO_BOX); // because it is filled with browser
    //bkmkwindow->border(0);
    //bkmkwindow->box(FL_UP_BOX);
    bkmkbrowser = new Fl_Paged_Select_Browser(0,0,Fl_Group::current()->w(), Fl_Group::current()->h()-25);
    bkmkbrowser->textfont(FL_HELVETICA);
    bkmkbrowser->textsize(8);
#else
    int w = (8*Fl::w())/10;
    int h = (8*Fl::h())/10;
    if ((14*h)/w < w) w = (10*h)/14;
    else h = (14*w)/10;
    bkmkwindow = new Fl_Window(w,h,"Choose Bookmark");
    //    bkmkwindow->box(FL_NO_BOX); // because it is filled with browser
    bkmkbrowser = new Fl_Select_Browser(0,0,Fl_Group::current()->w(), Fl_Group::current()->h()-25);
#endif
    bkmkbrowser->callback(bkmk_cb,this);
    Fl_Button *jump_accept = new Fl_Button((bkmkbrowser->w()-35)/2,bkmkbrowser->h()+5,35,15,"Cancel");
    jump_accept->labelsize(10);
    jump_accept->callback(Cancel_cb,NULL);
    bkmkwindow->end();
    bkmkwindow->set_modal();
  }

  static void bkmk_cb(Fl_Widget* o, void* _data)
  {
    o->window()->hide();
    int pos = ((Fl_Browser*)o)->value();
    //    printf("callback, selection = <%d:%s>, event_clicks = %d\n",
    //   pos, ((Fl_Browser*)o)->text(pos), Fl::event_clicks());
    if (pos != 0)
      {
	if (((reader_ui*)_data)->bkmkfind)
	  {
	    (((reader_ui*)_data)->input)->locate((unsigned int)((Fl_Browser*)o)->data(pos));
	  }
	else
	  {
	    if (((reader_ui*)_data)->pBkmklist != NULL)
	      {
		((reader_ui*)_data)->pBkmklist->erase(pos);
	      }
	    else
	      {
		DBT key,data;
		memset(&key,0,sizeof(key));
		memset(&data,0,sizeof(data));
		setbkmkkey(((reader_ui *)_data)->g_filename,((Fl_Browser*)o)->text(pos),&key);
		int ret = bkmks->get(bkmks, &key, &data, 0);
		if (ret == 0)
		  {
		    if (bkmks->del(bkmks, &key, 0) != 0)
		      {
			fl_alert("Problem deleting bookmark:%d",errno);
		      }
		  }
		else
		  {
		    fl_alert("Bookmark not found");
		  }
	      }
	  }
      }
    ((Fl_Browser*)o)->clear();
  }

  static void populate_bkmk(reader_ui *r)
  {
    CQueue<Bkmk>* t = r->pBkmklist;;
    r->bkmkbrowser->clear();
    int cnt = 0;
    if (t != NULL)
      {
	for (CList<Bkmk>::iterator i = t->begin(); i != t->end(); i++)
	  {
	    r->bkmkbrowser->add(i->name(), (void*)i->value());
	    cnt++;
	  }
      }
    else
      {
	DBT key,data;
	memset(&key,0,sizeof(key));
	memset(&data,0,sizeof(data));
	char buffer1[2048];
	strcpy(buffer1,r->g_filename);
	int offset = strlen(r->g_filename);
	key.data = buffer1;
	key.size = offset;
	bkmks->sync(bkmks, 0);
	CList<Bkmk>* t;
	if ((t = r->input->getbkmklist()) != NULL)
	  {
	    for (CList<Bkmk>::iterator i = t->begin(); i != t->end(); i++)
	      {
		r->bkmkbrowser->add(i->name(), (void*)i->value());
		cnt++;
	      }
	    delete t;
	  }
	if (bkmks->seq(bkmks,&key,&data,R_CURSOR) == 0)
	  {
	    CQueue<Bkmk> blist;
	    while (strcmp(r->g_filename,(char*)key.data)==0)
	      {
		char buffer[4096];
		memcpy(buffer, (char*)key.data+offset+1, key.size-offset-1);
		buffer[key.size-offset-1] = '\0';
		//printf("Found:%s-%s\n",buffer1,buffer);
		if (strcmp(buffer,"CURRENT") != 0)
		  {
		    unsigned long lcn;
		    memcpy(&lcn,data.data,sizeof(lcn));
		    //		    r->bkmkbrowser->add(buffer,(void*)lcn);
		    blist.push(Bkmk(buffer,lcn));
		    cnt++;
		  }
		if (bkmks->seq(bkmks,&key,&data,R_NEXT) != 0) break;
	      }
	    blist.sort();
	    while (!blist.empty())
	      {
		Bkmk b(blist.pop());
		r->bkmkbrowser->add(b.name(), (void*)b.value());
	      }
	  }
      }
    //    printf("Showing\n");
    //    r->bkmkbrowser->position(0);
    if (cnt > 0)
      r->bkmkwindow->show();
    else
      fl_alert("No bookmarks for\n%s",filename_name(r->g_filename));
  }

  static void getbkmk_cb(Fl_Widget* o, void* _data)
  {
    reader_ui* r = (reader_ui *)_data;
    r->bkmkfind = true;
    populate_bkmk(r);
  }

  static void delbkmk_cb(Fl_Widget* o, void* _data)
  {
    reader_ui* r = (reader_ui *)_data;
    r->bkmkfind = false;
    populate_bkmk(r);
  }
#endif

  static void bstripcr_cb(Fl_Widget* o, void* _data)
  {
    ((reader_ui*)_data)->bstripcr = !(((reader_ui*)_data)->bstripcr);
    ((reader_ui*)_data)->setfilter();
  }

  static void bunindent_cb(Fl_Widget* o, void* _data)
  {
    ((reader_ui*)_data)->bunindent = !(((reader_ui*)_data)->bunindent);
    ((reader_ui*)_data)->setfilter();
  }

  static void brepara_cb(Fl_Widget* o, void* _data)
  {
    ((reader_ui*)_data)->brepara = !(((reader_ui*)_data)->brepara);
    ((reader_ui*)_data)->setfilter();
  }

  static void bindenterp_cb(Fl_Widget* o, void* _data)
  {
    ((reader_ui*)_data)->bindenter += 2;
    ((reader_ui*)_data)->setfilter();
  }

  static void bindenterm_cb(Fl_Widget* o, void* _data)
  {
    if (((reader_ui*)_data)->bindenter < 2)
      ((reader_ui*)_data)->bindenter = 0;
    else
      ((reader_ui*)_data)->bindenter -= 2;
    ((reader_ui*)_data)->setfilter();
  }

  static void bdblspce_cb(Fl_Widget* o, void* _data)
  {
    ((reader_ui*)_data)->bdblspce = !(((reader_ui*)_data)->bdblspce);
    ((reader_ui*)_data)->setfilter();
  }

  static void info_cb(Fl_Widget* o, void* _data)
  {

    static Fl_Window* infowin = NULL;
    static Fl_Output* filename;
    static Fl_Output* filesize;
    static Fl_Output* textsize;
    static CBar* comprat;
    static Fl_Output* posn;
    static CBar* frcn;
    static Fl_Multiline_Output* about;
    if (infowin == NULL)
      {
	infowin = new Fl_Window(160,240);
	//infowin = new_cwindow(160,240);
	//infowin->border(0);
	//infowin->box(FL_UP_BOX);
	filename = new Fl_Output(45,5,110,14,"Filename");
	filename->textsize(10);
	filename->labelsize(10);

	filesize = new Fl_Output(45,25,110,14,"Filesize");
	filesize->textsize(10);
	filesize->labelsize(10);

	textsize = new Fl_Output(45,45,110,14,"Textsize");
	textsize->textsize(10);
	textsize->labelsize(10);

	comprat = new CBar(45,65,110,14,"Ratio %");
	comprat->labelsize(10);

	posn = new Fl_Output(45,85,110,14,"Location");
	posn->textsize(10);
	posn->labelsize(10);

	frcn = new CBar(45,105,110,14,"% Read");
	frcn->labelsize(10);

	about = new Fl_Multiline_Output(5,125,150,90);
	about->textsize(10);
	about->value("TWReader - $Name: V1_2 $\n\nA file reader program for the Agenda\n\nReads text, PalmDoc and ppms format files");

	Fl_Button *jump_accept = new Fl_Button(62,220,35,14,"Okay");
	jump_accept->labelsize(10);
	jump_accept->callback(Cancel_cb,_data);
	infowin->end();
	infowin->set_modal();
      }
    if (((reader_ui *)_data)->g_filename[0] != '\0')
      {
	unsigned long fs,ts;
	char sz[20];
	((reader_ui *)_data)->input->sizes(fs,ts);
	unsigned long pl = ((reader_ui *)_data)->input->locate();

	filename->value(((reader_ui *)_data)->g_filename);

	sprintf(sz,"%u",fs);
	filesize->value(sz);

	sprintf(sz,"%u",ts);
	textsize->value(sz);

	comprat->value(100-(100*fs + (ts >> 1))/ts);

	sprintf(sz,"%u",pl);
	posn->value(sz);

	frcn->value((100*pl + (ts >> 1))/ts);
      }
    infowin->show();
  }

  reader_ui() :
    bstripcr(false), bunindent(false), brepara(false), bindenter(0),
    bdblspce(false), btight(false), pBkmklist(NULL)
  {
    g_filename[0] = '\0';
#ifdef AGENDA
    window = Widget_Factory::new_window("TWReader");
    input = new myoutput(0, 0, Fl_Group::current()->w(), Fl_Group::current()->h());
    input->textfont(FL_HELVETICA);
    input->textsize(8);
    Fl_Group::current()->resizable(input);
    getprefs();
    input->tight = btight;
    int boolflag = FL_MENU_TOGGLE | ((bold) ? FL_MENU_VALUE : 0);
    int stripflag = FL_MENU_TOGGLE | ((bstripcr) ? FL_MENU_VALUE : 0);
    int unindentflag = FL_MENU_TOGGLE | ((bunindent) ? FL_MENU_VALUE : 0);
    int paraflag = FL_MENU_TOGGLE | ((brepara) ? FL_MENU_VALUE : 0);
    int dblflag = FL_MENU_TOGGLE | ((bdblspce) ? FL_MENU_VALUE : 0);
    int tightflag = FL_MENU_TOGGLE | ((btight) ? FL_MENU_VALUE : 0);

    Fl_Dockable_Window* toolbar = Widget_Factory::new_toolbar();
    Fl_Button *done_button = Widget_Factory::new_button("Done", quit_cb, this);
    Fl_Menu_Button* menubar = Widget_Factory::new_menu_button("File");
    menubar->add("Open",0,open_cb,this);
    menubar->add("Info",0,info_cb,this);
    menubar->add("Find",0,find_cb,this);
    menubar->add("Find Again",0,find2_cb,this);
    menubar->add("Jump",0,jump_cb,this);
#ifdef USEBKMKS
    Fl_Menu_Button* bkmkmb = menubar = Widget_Factory::new_menu_button("Marks");
    menubar->add("Mark",0,save_cb,this);
    menubar->add("Goto",0,getbkmk_cb,this);
    menubar->add("Delete",0,delbkmk_cb,this);
    menubar->add("Autogen",0,autogenbkmk_cb,this);
    menubar->add("Clear",0,clrbkmk_cb,this);
    menubar->add("Tidy",0,tidybkmk_cb,this);
    menubar->add("Save",0,savebkmk_cb,this,FL_MENU_INACTIVE);
    menubar->add("Start Block", 0, mark_cb, this);
    menubar->add("Copy", 0, copy_cb, this);
#endif
    Fl_Menu_Button* vmb = menubar = Widget_Factory::new_menu_button("View");
    menubar->add("Magnify",0,mag_cb,this);
    menubar->add("Reduce",0,red_cb,this);
    menubar->add("Font",0,fnt_cb,this);
    menubar->add("Bold",0,bld_cb,this,boolflag);
    menubar->add("Tight",0,tight_cb,this,tightflag);
    menubar = Widget_Factory::new_menu_button("Format");
    menubar->add("StripCR",0,bstripcr_cb,this,stripflag);
    menubar->add("Unindent",0,bunindent_cb,this,unindentflag);
    menubar->add("Repara",0,brepara_cb,this,paraflag);
    menubar->add("Indent+",0,bindenterp_cb,this,0);
    menubar->add("Indent-",0,bindenterm_cb,this,0);
    menubar->add("DblSpace",0,bdblspce_cb,this,dblflag);
    toolbar->end();
    window->add_dockable(toolbar,1);


    for (int j = 0; j < vmb->size()-1; j++)
      {
	if ((vmb->menu()+j)->callback() == red_cb)
	  mi_zoomout = const_cast<Fl_Menu_Item*>(vmb->menu()+j);
	if ((vmb->menu()+j)->callback() == mag_cb)
	  mi_zoomin = const_cast<Fl_Menu_Item*>(vmb->menu()+j);
      }

    {
      int * sizep;
      int n = Fl::get_font_sizes(input->textfont(),sizep);
      if (n != 0)
	{
	  if (input->textsize() == sizep[0] || (sizep[0] == 0 && input->textsize() == sizep[1]))
	    {
	      mi_zoomout->deactivate();
	    }
	  if (input->textsize() == sizep[n-1])
	    {
	      mi_zoomin->deactivate();
	    }

	}
    }

#ifdef USEBKMKS
    for (int j = 0; j < bkmkmb->size()-1; j++)
      {
	if ((bkmkmb->menu()+j)->callback() == savebkmk_cb)
	  mi_savebkmk = const_cast<Fl_Menu_Item*>(bkmkmb->menu()+j);
	if ((bkmkmb->menu()+j)->callback() == autogenbkmk_cb)
	  mi_autogen = const_cast<Fl_Menu_Item*>(bkmkmb->menu()+j);
      }
#endif

#else
    window = new Fl_Window(Fl::w()/2,Fl::h()/2,"TWReader");
#ifdef WS
    input = new myoutput(0, 20, window->w(), window->h()-20);
#else
    input = new myoutput(0, 12, window->w(), window->h()-12);
#endif
    input->textfont(FL_HELVETICA);
    input->textsize(12);
    window->resizable(input);
    getprefs();
    input->tight = btight;
    int boolflag = FL_MENU_TOGGLE | ((bold) ? FL_MENU_VALUE : 0);
    int stripflag = FL_MENU_TOGGLE | ((bstripcr) ? FL_MENU_VALUE : 0);
    int unindentflag = FL_MENU_TOGGLE | ((bunindent) ? FL_MENU_VALUE : 0);
    int paraflag = FL_MENU_TOGGLE | ((brepara) ? FL_MENU_VALUE : 0);
    int dblflag = FL_MENU_TOGGLE | ((bdblspce) ? FL_MENU_VALUE : 0);
    int tightflag = FL_MENU_TOGGLE | ((btight) ? FL_MENU_VALUE : 0);
    static Fl_Menu_Item menuitems[] =
    {
      { "File", 0, 0, 0, FL_SUBMENU },
      { "&Open", FL_ALT+'o', open_cb, this },
      { "&Quit", FL_ALT+'q', quit_cb, this},
      { "&Info", FL_ALT+'i', info_cb,this},
      { 0 },
      { "Search", 0, 0, 0, FL_SUBMENU },
      { "&Find",FL_ALT+'f',find_cb,this},
      {"Find A&gain",FL_ALT+'g',find2_cb,this},
      {"&Jump",FL_ALT+'j',jump_cb,this},
#ifdef USEBKMKS
      { "&Mark Current", FL_ALT+'m', save_cb,this},
      {"Goto mar&k",FL_ALT+'k',getbkmk_cb,this},
      { "More Bookmarks", 0, 0, 0, FL_SUBMENU },
      { "&Delete Mark", FL_ALT+'d', delbkmk_cb, this},
      {"&Autogen",FL_ALT+'a',autogenbkmk_cb,this},
      {"C&lear",FL_ALT+'l',clrbkmk_cb,this},
      {"&Tidy",FL_ALT+'t',tidybkmk_cb,this},
      {"&Save",FL_ALT+'s',savebkmk_cb,this,FL_MENU_INACTIVE},
      {"Block", FL_ALT+' ', mark_cb, this},
      {"&Copy", FL_ALT+'c', copy_cb, this},
      {0},
#endif
      {0},
      {"View",0,0,0,FL_SUBMENU},
      {"&Zoom in",FL_ALT+'z',mag_cb,this},
      {"&Zoom out",FL_ALT+'Z',red_cb,this},
      {"Fo&nt",FL_ALT+'n',fnt_cb,this},
      {"Bold",FL_ALT+'b',bld_cb,this,boolflag},
      {"Tight",FL_ALT+'w',tight_cb,this,tightflag},
      {0},
      {"Format",0,0,0,FL_SUBMENU},
      {"StripCR",0,bstripcr_cb,this,stripflag},
      {"Unindent",0,bunindent_cb,this,unindentflag},
      {"Repara",0,brepara_cb,this,paraflag},
      {"Indent+",0,bindenterp_cb,this,0},
      {"Indent-",0,bindenterm_cb,this,0},
      {"DblSpace",0,bdblspce_cb,this,dblflag},
      {0},
      {0}
    };
    int i;
    for (i = 0; menuitems[i].shortcut() != FL_ALT+'z'; i++);
    mi_zoomin = menuitems+i;
    for (i = 0; menuitems[i].shortcut() != FL_ALT+'Z'; i++);
    mi_zoomout = menuitems+i;
    for (i = 0; menuitems[i].shortcut() != FL_ALT+'s'; i++);
    mi_savebkmk = menuitems+i;
    for (i = 0; menuitems[i].shortcut() != FL_ALT+'a'; i++);
    mi_autogen = menuitems+i;
#ifdef WS
    Fl_Menu_Bar *menubar= new Fl_Menu_Bar(0,0,window->w(),20);
#else
    Fl_Menu_Bar *menubar= new Fl_Menu_Bar(0,0,window->w(),12);
    menubar->textsize(8);
#endif
    menubar->menu(menuitems);
    window->end();
#endif
    //    input = new myoutput(0, 0, Fl_Group::current()->w(), Fl_Group::current()->h(),toolbar);
    //
    //    Fl_Group::current()->resizable(input);
    //    input->callback((Fl_Callback *)changed_cb);
    //    input->when(FL_WHEN_CHANGED);
#ifdef USEBKMKS
    setupbkmks();
#endif
    //  window->callback((Fl_Callback *)quit_cb);
    setfilter();
    window->show();
  }

  void load_file(char *newfile) {
#ifdef USEBKMKS
    save_current(NULL,this);
    if (pBkmklist != NULL) delete pBkmklist;
    mark = 0;
#endif
    strcpy(g_filename,newfile);
#ifdef USEBKMKS
    if (input->load_file(newfile,initlocation(newfile)))
#else
    if (input->load_file(newfile,0))
#endif
      {
        window->label(filename_name(newfile));
      }
    else
      {
        window->label("TWReader");
      }
  }
  void locate(unsigned int lcn) { input->locate(lcn); }

  void getprefs()
  {
    prefs_t prefs;
#ifdef USEBKMKS
    DBT key, data;
    memset(&key,0,sizeof(key));
    memset(&data,0,sizeof(data));
    char buffer[] = PREFRNCE;
    key.data = buffer;
    key.size = strlen(buffer);
    int ret = bkmks->get(bkmks, &key, &data, 0);
    if (ret != 0 || data.size != sizeof(prefs))
      {
	//	fl_alert("Incompatible or missing preferences");
      }
    else
      {
	memcpy(&prefs,data.data,data.size);
	//printf("%d:%d:%d\n",prefs.textsize,prefs.textfont,prefs.bold);
	input->textsize(prefs.textsize);
	input->textfont(prefs.textfont);
	if (prefs.bold)
	  {
	    bold = true;
	  }
	bstripcr = prefs.bstripcr;
	bunindent = prefs.bunindent;
	brepara = prefs.brepara;
	bdblspce = prefs.bdblspce;
	btight = prefs.btight;
	bindenter = prefs.bindenter;
      }
#else
    FILE *f = fopen(".twr-bkmks","rb");
    if (f != NULL)
      {
	size_t n = fread(&prefs,sizeof(prefs),1,f);
	if (n == 1)
	  {
	    input->textsize(prefs.textsize);
	    input->textfont(prefs.textfont);
	    if (prefs.bold)
	      {
		bold = true;
	      }
	    bstripcr = prefs.bstripcr;
	    bunindent = prefs.bunindent;
	    brepara = prefs.brepara;
	    bdblspce = prefs.bdblspce;
	    btight = prefs.btight;
	    bindenter = prefs.bindenter;
	    size_t lcn;
	    n = fread(&lcn,sizeof(lcn),1,f);
	    if (n == 1)
	      {
		char filename[1024];
		memset(filename,0,1024);
		n = fread(filename,sizeof(char),1023,f);
		strcpy(g_filename,filename);
		if (n > 0)
		  {
		    setfilter();
		    input->load_file(filename,lcn);
		  }
	      }
	  }
	fclose(f);
      }
#endif
  }

  void setprefs()
  {
    prefs_t prefs;
    prefs.textsize = input->textsize();
    prefs.textfont = input->textfont();
    prefs.bold = bold;
    prefs.bstripcr = bstripcr;
    prefs.bunindent = bunindent;
    prefs.brepara = brepara;
    prefs.bdblspce = bdblspce;
    prefs.btight = btight;
    prefs.bindenter = bindenter;
    //printf("%d:%d:%d\n",prefs.textsize,prefs.textfont,prefs.bold);
#ifdef USEBKMKS
    DBT key, data;
    memset(&key,0,sizeof(key));
    memset(&data,0,sizeof(data));
    char buffer[] = PREFRNCE;
    key.data = buffer;
    key.size = strlen(buffer);
    //printf("%d:%d:%d\n",prefs.textsize,prefs.textfont,prefs.bold);
    data.data = &prefs;
    data.size = sizeof(prefs);
    int ret = bkmks->put(bkmks, &key, &data, 0);
#else
    FILE *f = fopen(".twr-bkmks","wb");
    fwrite(&prefs,sizeof(prefs),1,f);
    size_t lcn = input->pagelocate();
    if (g_filename[0] != '\0')
      {
	fwrite(&lcn,sizeof(lcn),1,f);
	fwrite(g_filename,1,strlen(g_filename),f);
      }
    fclose(f);
#endif
  }
};

#ifdef USEBKMKS
unsigned int reader_ui::initlocation(char *file) {
  DBT key, data;
  memset(&key,0,sizeof(key));
  memset(&data,0,sizeof(data));
  key.data = file;
  key.size = strlen(file);
  getdefaultbkmk(file,&key);
  int ret = bkmks->get(bkmks, &key, &data, 0);
  if (ret == 0)
    {
      unsigned long lcn;
      memcpy(&lcn,data.data,sizeof(lcn));
      return lcn;
    }
  else
    {
      return 0;
    }
}
#endif

int main(int argc, char **argv) {
#ifdef USEBKMKS
  bkmks = dbopen(".twr-bkmks", O_RDWR|O_CREAT, S_IRWXU, DB_BTREE, NULL);
  //  fl_message_font(FL_HELVETICA,8);
  if (bkmks == NULL)
    {
      fl_alert("ERROR: can't open bookmarks - the file may be corrupt.");
      return 10;
    }
#endif

  reader_ui reader;

  filename[0] = '\0';
  search[0] = '\0';

  if (argc > 1) {
    unsigned int lcn = 0;
    if (argc > 2) lcn = atoi(argv[2]);
    reader.load_file(argv[1]);
    
    if (argc > 2) reader.locate(lcn);
    
  }
  else
    {
#ifdef USEBKMKS
      DBT key, data;
      memset(&key,0,sizeof(key));
      memset(&data,0,sizeof(data));
      char newfile[1024];
      char buffer[1024];
      strcpy(buffer,LASTFILE);
      key.data = buffer;
      key.size = strlen(buffer);
      int ret = bkmks->get(bkmks, &key, &data, 0);
      if (ret == 0)
	{
	  memcpy(newfile,data.data,data.size);
	  newfile[data.size] = '\0';
	  reader.load_file(newfile);
	}
#endif
    }
  fl_message_font(FL_HELVETICA,8);
  return Fl::run();
}
//
// End of "$Id: reader.cxx,v 1.40 2001/11/07 21:33:58 agenda Exp $".
//
